import logging
from dataclasses import _MISSING_TYPE as DEFAULT_MISSING_TYPE
from dataclasses import Field as DataclassField
from dataclasses import dataclass, field
from dataclasses import fields as dataclass_fields
from dataclasses import is_dataclass
from typing import Any, Callable, Dict, Iterator, List, Optional, Type, TypeVar

import stringcase

from .compat import T, assert_type, is_dict, is_list, is_opt, is_tuple, is_union, type_args

logger = logging.getLogger('serde')

JsonValue = TypeVar('JsonValue', str, int, float, bool, Dict, List)

SE_NAME = '__serde_serialize__'

FROM_ITER = '__serde_from_iter__'

FROM_DICT = '__serde_from_dict__'

TO_ITER = '__serde_to_iter__'

TO_DICT = '__serde_to_dict__'

HIDDEN_NAME = '__serde_hidden__'

SETTINGS = dict(debug=False)


@dataclass
class Hidden:
    """
    Hidden infomation encoded in serde classes.
    """

    code: Dict[str, str] = field(default_factory=dict)


def init(debug: bool = False):
    SETTINGS['debug'] = debug


class SerdeError(TypeError):
    """
    Serde error class.
    """


def gen(code: str, globals: Dict = None, locals: Dict = None, cls: Type = None) -> str:
    """
    Customized `exec` function.
    """
    try:
        from black import format_str, FileMode

        code = format_str(code, mode=FileMode(line_length=100))
    except Exception:
        pass
    for_class = 'for ' + cls.__name__ if cls else ''
    logger.debug(f'Generating {for_class} ...\n{code}')
    exec(code, globals, locals)
    return code


def typecheck(cls: Type[T], obj: T) -> None:
    """
    type check type-annotated classes.

    >>> @dataclass
    ... class Foo:
    ...     s: str
    >>>
    >>> typecheck(Foo, Foo('foo'))
    >>>
    >>> # Type mismatch raises `ValueError`.
    >>> try:
    ...     typecheck(Foo, Foo(10))
    ... except:
    ...     pass
    >>>
    """
    if is_dataclass(obj):
        # If dataclass, type check recursively.
        for f in dataclass_fields(obj):
            typecheck(f.type, getattr(obj, f.name, None))
    elif is_opt(cls):
        if obj is not None:
            typ = type_args(cls)[0]
            typecheck(typ, obj)
    elif is_union(cls):
        success = False
        for typ in type_args(cls):
            try:
                typecheck(typ, obj)
                success = True
                break
            except (SerdeError, ValueError):
                pass
        if not success:
            raise ValueError(f'{obj} is not instance of {cls}')
    elif is_list(cls):
        assert_type(list, obj)
        if isinstance(obj, list):
            typ = type_args(cls)[0]
            for e in obj:
                typecheck(typ, e)
    elif is_tuple(cls):
        assert_type(tuple, obj)
        if isinstance(obj, tuple):
            for i, typ in enumerate(type_args(cls)):
                typecheck(typ, obj[i])
    elif is_dict(cls):
        assert_type(dict, obj)
        if isinstance(obj, dict):
            ktyp = type_args(cls)[0]
            vtyp = type_args(cls)[1]
            for k, v in obj.items():
                typecheck(ktyp, k)
                typecheck(vtyp, v)
    else:
        if not isinstance(obj, cls):
            raise ValueError(f'{obj} is not instance of {cls}')


@dataclass
class Field:
    """
    Field in pyserde class.
    """

    type: Type
    name: str
    default: Any = field(default_factory=DEFAULT_MISSING_TYPE)
    case: Optional[str] = None
    rename: Optional[str] = None
    skip: Optional[bool] = None
    skip_if: Optional[Callable[[Any], bool]] = None
    skip_if_false: Optional[bool] = None

    @classmethod
    def from_dataclass(cls, f: DataclassField) -> 'Field':
        if f.metadata.get('serde_skip_if_false'):

            def skip_if_false(v):
                return not bool(v)

            skip_if_false.mangled = cls.mangle(f, 'skip_if')
        else:
            skip_if_false = None

        skip_if = f.metadata.get('serde_skip_if')
        if skip_if:
            skip_if.mangled = cls.mangle(f, 'skip_if')

        return cls(
            f.type,
            f.name,
            default=f.default,
            rename=f.metadata.get('serde_rename'),
            skip=f.metadata.get('serde_skip'),
            skip_if=skip_if or skip_if_false,
        )


def fields(FieldCls: Type, cls: Type) -> Iterator[Field]:
    return iter(FieldCls.from_dataclass(f) for f in dataclass_fields(cls))


def conv(f: Field, case: Optional[str] = None) -> str:
    """
    Convert field name.
    """
    name = f.name
    if case:
        casef = getattr(stringcase, case or '', None)
        if not casef:
            raise SerdeError(
                (f"Unkown case type: {f.case}." f"Pass the name of case supported by 'stringcase' package.")
            )
        name = casef(f.name)
    if f.rename:
        name = f.rename
    return name



from typing import Any, Type

import msgpack

from .core import T, logger
from .de import Deserializer, from_dict, from_tuple
from .se import Serializer, asdict, astuple


class MsgPackSerializer(Serializer):
    @classmethod
    def serialize(cls, obj, named=True, **opts) -> str:
        asf = asdict if named else astuple
        return msgpack.packb(asf(obj), use_bin_type=True, **opts)


class MsgPackDeserializer(Deserializer):
    @classmethod
    def deserialize(cls, s, **opts):
        unp = msgpack.unpackb(s, raw=False, use_list=False, **opts)
        logger.debug('unpack from msgpack: {unp}')
        return unp


def to_msgpack(obj: Any, serializer=MsgPackSerializer, named=True, **opts) -> bytes:
    return obj.__serde_serialize__(serializer, **opts)


def from_msgpack(c: Type[T], s: str, de: Type[Deserializer] = MsgPackDeserializer, named=True, **opts) -> Type[T]:
    fromf = from_dict if named else from_tuple
    return fromf(c, de.deserialize(s, **opts))



"""
pyserde inspect tool.

USAGE:

$ python -m serde.inspect <PATH> <NAME>

 PATH  Python script path.
 NAME  Pyserde class name.

"""
import argparse
import importlib
import os
import sys
from typing import Type

from .core import HIDDEN_NAME, init

init(True)


def inspect(cls: Type) -> str:
    """
    Inspect a pyserde class.
    """
    hidden = getattr(cls, HIDDEN_NAME, None)
    return '\n'.join(hidden.code.values())


def main(arg):
    dir = os.path.dirname(arg.path)
    mod = os.path.basename(arg.path)[:-3]
    print(f'Loading {mod}.{arg.name} from {dir}.')
    sys.path.append(dir)
    pkg = importlib.import_module(mod)
    cls = getattr(pkg, arg.name)
    print(inspect(cls))


parser = argparse.ArgumentParser(description='pyserde-inspect')
parser.add_argument('path', type=str, help='Python script path.')
parser.add_argument('name', type=str, help='Pyserde class name.')


if __name__ == '__main__':
    args = parser.parse_args()
    main(args)



import json
from typing import Any, Type

from .core import T
from .de import Deserializer, from_dict
from .se import Serializer, asdict


class JsonSerializer(Serializer):
    @classmethod
    def serialize(cls, obj: Any, named=True, **opts) -> str:
        return json.dumps(asdict(obj), **opts)


class JsonDeserializer(Deserializer):
    @classmethod
    def deserialize(cls, s, named=True, **opts):
        return json.loads(s, **opts)


def to_json(obj: Any, cls: Type[JsonSerializer] = JsonSerializer, **opts) -> str:
    return cls.serialize(obj)


def from_json(c: Type[T], s: str, de: Type[Deserializer] = JsonDeserializer, **opts) -> T:
    return from_dict(c, de.deserialize(s, **opts))



from dataclasses import dataclass, field
from serde import deserialize, serialize
from serde.json import from_json, to_json


@deserialize
@serialize
@dataclass
class Foo:
    i: int = 10
    s: str = 'foo'
    f: float = field(default=100.0)  # Use dataclass field.
    b: bool = field(default=True)


def main():
    h = Foo()
    print(f"Into Json: {to_json(h)}")

    s = '{"i": 10, "s": "foo", "f": 100.0, "b": true}'
    print(f"From Json: {from_json(Foo, s)}")


if __name__ == '__main__':
    main()



import sys

import basics
import default
import env
import jsonfile
import rename
import skip
import tomlfile
import yamlfile


def run_all():
    run(basics)
    run(default)
    run(env)
    run(rename)
    run(skip)
    run(jsonfile)
    run(tomlfile)
    run(yamlfile)


def run(module):
    print(f'running {module.__name__}')
    module.main()


if __name__ == '__main__':
    try:
        run_all()
    except Exception:
        sys.exit(1)



from dataclasses import fields, is_dataclass
from itertools import zip_longest
from typing import Iterator, Optional, Tuple, Type, TypeVar, Union

from typing_inspect import get_args, get_origin, is_optional_type, is_union_type

T = TypeVar('T')


def typename(typ) -> str:
    """
    >>> from typing import List, Dict
    >>> typename(int)
    'int'
    >>> class Foo: pass
    >>> typename(Foo)
    'Foo'
    >>> typename(List[Foo])
    'List[Foo]'
    >>> typename(Dict[str, Foo])
    'Dict[str, Foo]'
    >>> typename(Tuple[int, str, Foo, List[int], Dict[str, Foo]])
    'Tuple[int, str, Foo, List[int], Dict[str, Foo]]'
    >>> typename(Optional[List[Foo]])
    'Optional[List[Foo]]'
    >>> typename(Union[Optional[Foo], List[Foo], Union[str, int]])
    'Union[Optional[Foo], List[Foo], str, int]'
    """
    if is_opt(typ):
        return f'Optional[{typename(type_args(typ)[0])}]'
    elif is_union(typ):
        return f'Union[{", ".join([typename(e) for e in union_args(typ)])}]'
    elif is_list(typ):
        et = typename(type_args(typ)[0])
        return f'List[{et}]'
    elif is_dict(typ):
        kt = typename(type_args(typ)[0])
        vt = typename(type_args(typ)[1])
        return f'Dict[{kt}, {vt}]'
    elif is_tuple(typ):
        return f'Tuple[{", ".join([typename(e) for e in type_args(typ)])}]'
    else:
        return typ.__name__


def type_args(cls: Type):
    """
    Wrapepr to suppress type error for accessing private members.
    """
    return cls.__args__  # type: ignore


def union_args(typ: Union) -> Tuple:
    if not is_union(typ):
        raise TypeError(f'{typ} is not Union')
    args = type_args(typ)
    if len(args) == 1:
        return args[0]
    it = iter(args)
    types = []
    for (i1, i2) in zip_longest(it, it):
        if not i2:
            types.append(i1)
        elif is_none(i2):
            types.append(Optional[i1])
        else:
            types.extend((i1, i2))
    return tuple(types)


def iter_types(cls: Type) -> Iterator[Type]:
    """
    Iterate field types recursively.
    """
    if is_dataclass(cls):
        yield cls
        for f in fields(cls):
            yield from iter_types(f.type)
    elif isinstance(cls, str):
        yield cls
    elif is_opt(cls):
        yield from iter_types(type_args(cls)[0])
    elif is_union(cls):
        for arg in type_args(cls):
            yield from iter_types(arg)
    elif is_list(cls):
        yield from iter_types(type_args(cls)[0])
    elif is_tuple(cls):
        for arg in type_args(cls):
            yield from iter_types(arg)
    elif is_dict(cls):
        yield from iter_types(type_args(cls)[0])
        yield from iter_types(type_args(cls)[1])
    else:
        yield cls


def is_union(typ: Type) -> bool:
    """
    Test if the type is `typing.Union`.
    """
    return is_union_type(typ) and not is_opt(typ)


def is_opt(typ: Type) -> bool:
    """
    Test if the type is `typing.Optional`.
    """
    args = get_args(typ)
    return is_optional_type(typ) and len(args) == 2 and not is_none(args[0]) and is_none(args[1])


def is_list(typ: Type) -> bool:
    """
    Test if the type is `typing.List`.
    """
    try:
        return issubclass(get_origin(typ), list)
    except TypeError:
        return isinstance(typ, list)


def is_tuple(typ: Type) -> bool:
    """
    Test if the type is `typing.Tuple`.
    """
    try:
        return issubclass(get_origin(typ), tuple)
    except TypeError:
        return isinstance(typ, tuple)


def is_dict(typ: Type) -> bool:
    """
    Test if the type is `typing.Dict`.
    """
    try:
        return issubclass(get_origin(typ), dict)
    except TypeError:
        return isinstance(typ, dict)


def is_none(typ: Type) -> bool:
    """
    >>> is_none(int)
    False
    >>> is_none(type(None))
    True
    >>> is_none(None)
    False
    """
    return typ is type(None)  # noqa


PRIMITIVES = [int, float, bool, str]


def is_primitive(typ: Type) -> bool:
    """
    Test if the type is primitive.

    >>> is_primitive(int)
    True
    >>> is_primitive(float)
    True
    >>> class CustomInt(int):
    ...     pass
    >>> is_primitive(CustomInt)
    True
    """
    try:
        return any(issubclass(typ, ty) for ty in PRIMITIVES)
    except TypeError:
        return any(isinstance(typ, ty) for ty in PRIMITIVES)


def assert_type(typ: Type, obj, throw=False) -> None:
    if not isinstance(obj, typ):
        if throw:
            raise ValueError(f'{obj} is not instance of {typ}')



from dataclasses import dataclass
from decimal import Decimal

import pytest

from serde import deserialize, serialize
from tests.test_basics import format_dict, format_json, format_toml, format_yaml


@deserialize
@serialize
@dataclass
class Foo:
    d: Decimal


@pytest.mark.parametrize('se,de', format_json + format_yaml + format_toml)  # type: ignore
def test_decimal(se, de):
    f = Foo(d=Decimal('0.1'))
    assert f == de(Foo, se(f))

    f = Foo(d=Decimal(0.1))
    assert f == de(Foo, se(f))



"""
.. include:: ../README.md
"""
from .core import SerdeError, init, logger, typecheck  # noqa
from .de import deserialize, from_dict, from_obj, from_tuple, is_deserializable  # noqa
from .se import asdict, astuple, is_serializable, serialize  # noqa

""" Version of pyserde. """
__version__ = '0.0.11'



"""
csvfile.py

Read / Write USA city data in csv.

Usage:
    $ pipenv install
    $ pipenv run csvfile.py
"""
from dataclasses import dataclass
from serde import deserialize
from serde.csv import to_csv


@deserialize
@dataclass(unsafe_hash=True)
class City:
    city: str
    state: str
    population: int
    latitude: float
    longitude: float


def main():
    city = City('Davidsons Landing', 'AK', 0, 65.2419444, -165.2716667)
    print(to_csv(city))

    cities = [
        City('Davidsons Landing', 'AK', 0, 65.2419444, -165.2716667),
        City('Kenai', 'AK', 7610, 60.5544444, -151.2583333),
        City('Oakman', 'AL', 0, 33.7133333, -87.3886111),
        City('Richards Crossroads', 'AL', 0, 31.7369444, -85.2644444),
        City('Sandfort', 'AL', 0, 32.3380556, -85.2233333),
    ]
    print(to_csv(cities, header=True))


if __name__ == '__main__':
    main()



import io
import json
from csv import DictWriter as Writer
from dataclasses import fields
from typing import Any, List, Optional, Type, Union

from .core import T
from .de import Deserializer as De
from .de import from_dict
from .se import Serializer as Se
from .se import asdict


class CsvSerializer(Se):
    writer: Writer

    @classmethod
    def serialize(cls, obj: Union[T, List[T]], header: bool = False, **opts) -> Optional[io.StringIO]:
        if not obj:
            return None

        buf = io.StringIO()
        if isinstance(obj, list):
            head = obj[0]
            objs = obj
        else:
            head = obj
            objs = [obj]
        writer = Writer(buf, [f.name for f in fields(head)])

        if header:
            writer.writeheader()
        for obj in objs:
            writer.writerow(asdict(obj))
        buf.seek(0)
        return buf


class CsvDeserializer(De):
    @classmethod
    def deserialize(cls, s, **opts):
        return json.loads(s, **opts)


def to_csv(obj: Any, se: Type[Se] = CsvSerializer, **opts) -> str:
    res = se.serialize(obj, **opts)
    if res:
        return res.read().rstrip() or ''
    else:
        return ''


def from_csv(c: Type[T], s: str, de: Type[De] = CsvDeserializer, **opts) -> T:
    return from_dict(c, de.deserialize(s, **opts))



from dataclasses import Field
from decimal import Decimal
from typing import Any

from .core import SerdeError


def serialize(data: Any) -> Any:
    """
    Custom serializer for extended types such as Decimal.

    `data` shall be other than List, Dict, Tuple, Optional
    Union, dataclass or primitives.
    """
    if isinstance(data, Decimal):
        return str(data)
    else:
        raise SerdeError(f'Unsupported type: {type(data)}')


def deserialize(field: Field, data: Any) -> Any:
    """
    Custom deserializer for extended types such as Decimal.

    `data` shall be other than List, Dict, Tuple, Optional
    Union, dataclass or primitives.
    """
    if issubclass(field.type, Decimal):
        return Decimal(data)
    else:
        raise SerdeError(f'Unsupported type: {type(data)}')



from dataclasses import dataclass
from typing import Dict, List, Optional, Tuple, Union

import pytest

from serde import typecheck
from serde.compat import is_dict, is_list, is_opt, is_tuple, is_union, iter_types, type_args, union_args

from .data import Bool, Float, Int, Pri, PriOpt, Str


def test_types():
    assert is_list(List[int])
    assert is_tuple(Tuple[int, int, int])
    assert is_dict(Dict[str, int])
    assert is_opt(Optional[int])
    assert is_union(Union[int, str])
    assert is_union(Union[Optional[int], Optional[str]])
    assert is_opt(Optional[int])
    assert not is_union(Optional[int])
    assert not is_opt(Union[Optional[int], Optional[str]])
    assert is_union(Union[Optional[int], Optional[str]])


def test_iter_types():
    assert [Pri, int, str, float, bool] == list(iter_types(Pri))
    assert [str, Pri, int, str, float, bool] == list(iter_types(Dict[str, Pri]))
    assert [str] == list(iter_types(List[str]))
    assert [int, str, bool, float] == list(iter_types(Tuple[int, str, bool, float]))
    assert [PriOpt, int, str, float, bool] == list(iter_types(PriOpt))
    assert [Pri, int, str, float, bool, Pri, int, str, float, bool, Pri, int, str, float, bool] == list(
        iter_types(Tuple[List[Pri], Tuple[Pri, Pri]])
    )


def test_type_args():
    assert (int,) == type_args(List[int])
    assert (int, str) == type_args(Dict[int, str])
    assert (int, type(None)) == type_args(Optional[int])
    assert (Optional[int],) == type_args(List[Optional[int]])
    assert (List[int], type(None)) == type_args(Optional[List[int]])
    assert (List[int], Dict[str, int]) == type_args(Union[List[int], Dict[str, int]])
    assert (int, type(None), str) == type_args(Union[Optional[int], str])


def test_union_args():
    assert (int, str) == union_args(Union[int, str])
    assert (List[int], Dict[int, str]) == union_args(Union[List[int], Dict[int, str]])
    assert (Optional[int], str) == union_args(Union[Optional[int], str])


def test_typecheck():
    # Primitive
    typecheck(int, 10)
    with pytest.raises(ValueError):
        typecheck(int, 10.0)

    # Dataclass
    p = Pri(i=10, s='foo', f=100.0, b=True)
    typecheck(Pri, p)

    with pytest.raises(ValueError):
        p.i = 10.0
        typecheck(Pri, p)

    # Dataclass (Nested)
    @dataclass
    class Foo:
        p: Pri

    h = Foo(Pri(i=10, s='foo', f=100.0, b=True))
    typecheck(Foo, h)

    with pytest.raises(ValueError):
        h.p.i = 10.0
        typecheck(Foo, h)

    # List
    typecheck(List[int], [10])
    with pytest.raises(ValueError):
        typecheck(List[int], [10.0])

    # List of dataclasses
    typecheck(List[Int], [Int(n) for n in range(1, 10)])
    with pytest.raises(ValueError):
        typecheck(List[Pri], [Pri(i=10.0, s='foo', f=100.0, b=True)])

    # Tuple
    typecheck(Tuple[int, str, float, bool], (10, 'foo', 100.0, True))
    with pytest.raises(ValueError):
        typecheck(Tuple[int, str, float, bool], (10.0, 'foo', 100.0, True))

    # Tuple of dataclasses
    typecheck(Tuple[Int, Str, Float, Bool], (Int(10), Str('foo'), Float(100.0), Bool(True)))
    with pytest.raises(ValueError):
        typecheck(Tuple[Int, Str, Float, Bool], (Int(10.0), Str('foo'), Float(100.0), Bool(True)))

    # Dict
    typecheck(Dict[str, int], dict(foo=10, bar=20))
    with pytest.raises(ValueError):
        typecheck(Dict[str, int], dict(foo=10.0, bar=20))

    # Dict of dataclasses
    typecheck(Dict[Str, Int], {Str('foo'): Int(10), Str('bar'): Int(20)})
    with pytest.raises(ValueError):
        typecheck(Dict[Str, Int], {Str('foo'): Int(10.0), Str('bar'): Int(20)})

    # Optional
    typecheck(Optional[int], 10)
    typecheck(Optional[int], None)
    with pytest.raises(ValueError):
        typecheck(Optional[int], 10.0)

    # Optional of dataclass
    typecheck(Optional[Int], Int(10))
    typecheck(Optional[Int], None)
    with pytest.raises(ValueError):
        typecheck(Optional[Int], 10.0)
        typecheck(Optional[Int], Int(10.0))
